/*---------------------------------------------------------------------------*/
/*  YANOS - Yet Another Narrow Operating System               20070225 UdoP  */
/*-------------------------------------------------------------------------->>>
                                                              end of comment */
#include "extern.h"
#include "yanos.h"

/*---------------------------------------------------------------------------*/
/*  OS Definitions                                                           */
/*-------------------------------------------------------------------------->>>
Some definitions to make the code better readable.
OS_Schedule is a little bit tricky. OS_Scheduler (with r) will called every ms 
from a low priority timer interrupt routine. Additionally an OS_xyz() routine 
can generate the same interrupt like the hardware timer to force a rescheduling.
                                                              end of comment */
#define OS_EnterCriticalRegion asm ("fclr i")
#define OS_LeaveCriticalRegion asm ("fset i")
#define OS_Schedule() asm("int #23")
#define OS_LockSema(sema) OS_Sema|=sema
#define OS_UnlockSema(sema) OS_Sema&=~sema
#define OS_SemaAlreadyLocked OS_Sema&sema
#define OS_SignalAlreadySet OS_Signal&sig
#define OS_ClearSignal OS_Signal&=~sig
#define OS_IDLE 0

#define OS_CREATE_TASK(NAME,STACK)\
  U8 near os_stk_##NAME[STACK+0x20];\
  extern void NAME(void);

#define OS_INIT_TASK(ID,NAME)\
  os_stackpointer[ID]=(U32)os_stk_##NAME+sizeof(os_stk_##NAME)-30;\
  (* (volatile U32 *) os_stackpointer[ID])    =(U32)NAME;\
  (*((volatile U16 *)(os_stackpointer[ID])+2))=0x00c0;

/*---------------------------------------------------------------------------*/
/*  OS Variables                                                             */
/*-------------------------------------------------------------------------->>>
With some restrictions one can save a lot of RAM. Maybe a port to the
Fujitsus consists only out of three friendly tasks (without Signal/Delay/Sema)
then the OS needs only 13 Bytes (plus space for context).
                                                              end of comment */
                                                              
const U8 far ContextTable[2]={0xff,0x18};     // store all, sp correction is 24
U16 near task_ctx_num;                        // Yeah - my first constant 
                                              // variable - allways zero

U16 near OS_DelayTimer[TASK_NUM];             // timer unnecessary for idle
U16 near OS_WaitForSema[TASK_NUM];            // also unnecessary - but
U16 near OS_WaitForSignal[TASK_NUM];          // easier to code and even faster
U16 near OS_TaskFriendlyCnt[TASK_NUM];        // same again
U16 near OS_TaskIsFriendly[TASK_NUM];         // same again

U16 near OS_Signal;                           // Signal Container
U16 near OS_Sema;                             // Sema Container  

U16 near os_curtsk;                           // the current task
U32 near os_stackpointer[TASK_NUM];           // TBD U16 ???

/*---------------------------------------------------------------------------*/
/*  OS Initialisation                                                        */
/*-------------------------------------------------------------------------->>>
Let the preprocessor work
                                                              end of comment */
OS_CREATE_ALL_TASKS;

void OS_Init(void) {
  OS_INIT_ALL_TASKS;
}

/*---------------------------------------------------------------------------*/
/*  OS delay timer handling                                                  */
/*-------------------------------------------------------------------------->>>
This function must called every millisecond by a realtime task. We use the
high priored timer interrupt tsk_1ms().
                                                              end of comment */
                                                              
void OS_TimerDec(void) {
  U16 i;
  for (i=1; i<TASK_NUM; i++) {                // no timer for idle
    if (OS_DelayTimer[i]) OS_DelayTimer[i]--; // decrement if running
  }
}                                                              
                                                              
/*---------------------------------------------------------------------------*/
/*  OS_Delay                                                                 */
/*-------------------------------------------------------------------------->>>
Stop the task for the given timer ticks. After that the task take again part on
round robin.
                                                              end of comment */
void OS_TaskDelay(U16 delay) {
  OS_EnterCriticalRegion;
  OS_DelayTimer[os_curtsk]=delay;
  OS_Schedule();
  OS_LeaveCriticalRegion;
}

/*---------------------------------------------------------------------------*/
/*  OS_Friendly                                                              */
/*-------------------------------------------------------------------------->>>
We do not have any priority. But a task can decide by itself to be friendly
to the other tasks. The FriendlyLevel is the quantity of own timeslots which 
the task voluntary leave for the others.
                                                              end of comment */
void OS_TaskSetFriendlyLevel(U16 level) {
  OS_EnterCriticalRegion;
  OS_TaskIsFriendly[os_curtsk]=level;
  OS_TaskFriendlyCnt[os_curtsk]=level;
  OS_Schedule();
  OS_LeaveCriticalRegion;
}

/*---------------------------------------------------------------------------*/
/*  OS_Signal                                                                */
/*-------------------------------------------------------------------------->>>
Normally a signal has one producer and one consumer. In special cases (e.g. the
nv-ram handling) a signal is handled like a request and can have more than one
producer. This is easier as implementing a counting semaphore. The code is not 
prepared to have more than one consumer
                                                              end of comment */
void OS_SetSignal(U16 sig) {
  U16 i,tsk;
  OS_EnterCriticalRegion;
  tsk=os_curtsk;                              // point to the current task
  for (i=OS_IDLE+1; i<TASK_NUM; i++) {        
    if (++tsk>TASK_NUM-1) {                   // go the round beginning with the
      tsk=OS_IDLE+1;                          // following task but without idle
    }
    if (OS_WaitForSignal[tsk]&sig) {          // if someone is waiting for this signal
      OS_WaitForSignal[tsk]&=~sig;            // let this task next time run
      OS_LeaveCriticalRegion;
      return;                                 // do this only for one
    }
  }
  OS_Signal|=sig;                             // else produce the signal
  OS_LeaveCriticalRegion;
}

void OS_WaitSignal(U16 sig) {
  OS_EnterCriticalRegion;
  if (OS_SignalAlreadySet) {                  // if the signal is already produced
    OS_ClearSignal;                           // than consume it
    OS_LeaveCriticalRegion;                   // 
    return;                                   // we need not to wait
  }
  OS_WaitForSignal[os_curtsk]|=sig;           // else wait for the signal
  OS_Schedule();                              // be friendly to the next
  OS_LeaveCriticalRegion;
}

U16 OS_WaitSignalWithTimeout(U16 sig, S16 tout) {
  S16 timeout;                                // TBD das ist noch nicht getestet
  OS_EnterCriticalRegion;
  if (OS_SignalAlreadySet) {                  // if the signal is already produced
    OS_ClearSignal;                           // than consume it
    OS_LeaveCriticalRegion;                   // 
    return OS_SIGNAL_GET_DIRECT;              // we need not to wait
  }
  timeout=TimeMS+tout;                        // do this without disturbance
  OS_LeaveCriticalRegion;                     // now we can leave critical

  while ((timeout-TimeMS) > 0) {              // do wait with timeout
    OS_TaskDelay(1);                          // capsuled be itself
    OS_EnterCriticalRegion;
    if (OS_SignalAlreadySet) {                // if produced in the meantime
      OS_ClearSignal;                         // than consume it
      OS_LeaveCriticalRegion;                 // 
      return OS_SIGNAL_GET_IN_TIME;           // get the signal within time
    }
    OS_LeaveCriticalRegion;                   // no signal till yet
  }
  return OS_SIGNAL_FAILED;                    // time proceeded without signal
}

/*---------------------------------------------------------------------------*/
/*  OS_Sema                                                                  */
/*-------------------------------------------------------------------------->>>
                                                              end of comment */
void OS_GetSema(U16 sema) {
  OS_EnterCriticalRegion;
  if (OS_SemaAlreadyLocked) {                 // if the sema is locked by another
    OS_WaitForSema[os_curtsk]|=sema;          // then wait for it
    OS_Schedule();                            // be friendly
    OS_LeaveCriticalRegion;
    return;
  } else {                                    // if the sema is free
    OS_LockSema(sema);                        // lock it
    OS_LeaveCriticalRegion;                   // and return
  }
}

void OS_ReleaseSema(U16 sema) {
  U16 i,tsk;
  OS_EnterCriticalRegion;
  tsk=os_curtsk;                              // point to the current task      
  for (i=OS_IDLE+1; i<TASK_NUM; i++) {                                           
    if (++tsk>TASK_NUM-1) {                   // go the round beginning with the
      tsk=OS_IDLE+1;                          // following task but without idle
    }
    if (OS_WaitForSema[tsk]&sema) {           // if another is waiting for the
      OS_WaitForSema[tsk]&=~sema;             // sema we just release make this
      OS_LeaveCriticalRegion;                 // task runable
      return;                                 // return - but keep sema locked
    }
  }
  OS_UnlockSema(sema);                        // no one is waiting - unlock and
  OS_LeaveCriticalRegion;                     // return
}
   
U16 OS_SemaInquiry(U16 sema) {
  OS_EnterCriticalRegion;
  if (OS_SemaAlreadyLocked) {                 // if the sema is locked
    OS_LeaveCriticalRegion;                   // return that it is locked
    return OS_SEMA_IS_LOCKED;
  } else {                                    // if the sema is free
    OS_LockSema(sema);                        // than lock it
    OS_LeaveCriticalRegion;
    return OS_SEMA_WAS_FREE;                  // and return that we get it
  }
}

  
/*---------------------------------------------------------------------------*/
/*  OS GetNextTask                                                           */
/*-------------------------------------------------------------------------->>>
                                                              end of comment */

void OS_GetNextTask(void) {
  U16 i;
  StartLoopFromCurtsk:
  for (i=OS_IDLE+1; i<TASK_NUM; i++) {        // check max each task beginning
    if (++os_curtsk>TASK_NUM-1) {             // with the next in the round
      os_curtsk=OS_IDLE+1;                    // but without idle 
    }
    if ((!OS_DelayTimer[os_curtsk])   &&      // if the task is not delayed
        (!OS_WaitForSema[os_curtsk])  &&      // and not waiting for a
        (!OS_WaitForSignal[os_curtsk])) {     // sema or a signal
      if (OS_TaskIsFriendly[os_curtsk]) {     // but it is a friendly task
        if (0==OS_TaskFriendlyCnt[os_curtsk]--) { // if friendly time run down
          OS_TaskFriendlyCnt[os_curtsk]=OS_TaskIsFriendly[os_curtsk];//reload
          return;                             // Friendly task get the slot
        } else {                              // else Friendly stay friendly
          goto StartLoopFromCurtsk;           // search again the next
        }
      }                                       // not (delayed||waiting||friendly)
      return;                                 // then next one was found
    }
  }
  os_curtsk=OS_IDLE;                          // there is nothing to do
}

/*---------------------------------------------------------------------------*/
/*  OS_Scheduler                                                             */
/*-------------------------------------------------------------------------->>>
The old TCB is optimized away. With the new scheduling algorithm we need only
save the context from the interrupted task. This will be done on the tasks 
own stack.
                                                              end of comment */
#pragma ASM
  	.align
  	.glb	_OS_Scheduler
  _OS_Scheduler:
    fset u                                     ; switch to user stack
    stctx _task_ctx_num,_ContextTable          ; save the context
    fclr u                                     ; back to interrupt stack
    popm R0,R1,R2                              ; get programcounter and flags
    fset u                                     ; again the user stack
    pushm R0,R1,R2                             ; push pc and flg to the task stack
    mov.w _os_curtsk:16,A0                     ; get curtsk
    shl.w #2,A0                                ; get os_stackpointer[curtsk]
    stc	sp,_os_stackpointer:16[A0]             ; save sp from the current task

    jsr	_OS_GetNextTask                        ; get the next task

    mov.w _os_curtsk:16,A0                     ; get the new curtsk (maybe the same)
    shl.w #2,A0                                ; get os_stackpointer[curtsk]
    ldc _os_stackpointer:16[A0],sp             ; load the new stackpointer
    popm R0,R1,R2                              ; get pc and flags from the last time
    fclr u                                     ; switch to interrupt stack
    pushm R0,R1,R2                             ; restore the interrupt stack
    fset u                                     ; back to user stack
    ldctx _task_ctx_num,_ContextTable          ; load the old context
    fclr u                                     ; all work is done
    reit                                       ; return from interrupt
#pragma ENDASM


